from functools import partial

import jax
from jax import random
from jax import nn
import jax.numpy as np
import jmp

import haiku as hk
from haiku import initializers
from einops import rearrange, repeat

from progen_transformer.utils import exists

# constants

ATTN_MASK_VALUE = -1e10

# helpers

LayerNorm = partial(hk.LayerNorm, create_scale = True, create_offset = False, axis = -1)

def fixed_pos_embedding(seq, dim):
    inv_freq = 1.0 / (10000 ** (np.arange(0, dim, 2) / dim))
    sinusoid_inp = np.einsum("i , j -> i j", np.arange(seq), inv_freq)
    sinusoid_inp = repeat(sinusoid_inp, "b n -> b (n r)", r = 2)[None, :, :]
    return np.sin(sinusoid_inp), np.cos(sinusoid_inp)

def rotate_every_two(x):
    x = rearrange(x, '... (d r) -> ... d r', r = 2)
    x1, x2 = x[..., 0], x[..., 1]
    x = np.stack((-x2, x1), axis=-1)
    return rearrange(x, "... d r -> ... (d r)")

def apply_rotary_pos_emb(x, sincos):
    sin, cos = sincos
    rot_dim = sin.shape[-1]
    x, x_pass = x[..., :rot_dim], x[..., rot_dim:]
    x = (x * cos) + (rotate_every_two(x) * sin)
    return np.concatenate((x, x_pass), axis = -1)

# classes

class LocalAttention(hk.Module):
    def __init__(
        self,
        *,
        name,
        dim,
        window_size,
        heads = 8,
        dim_head = 64
    ):
        super().__init__(name = name)
        self.heads = heads
        self.scale = dim_head ** -0.5
        self.window_size = window_size
        inner_dim = dim_head * heads

        self.norm = LayerNorm()
        self.to_qkv = hk.Linear(inner_dim * 3, with_bias = False)
        self.to_out = hk.Linear(dim)

    def __call__(self, x, *, pos_emb):
        x = self.norm(x)

        n, h, wsz = x.shape[0], self.heads, self.window_size
        assert (n % wsz) == 0, 'sequence length must be divisible by the window size'
        window = n // wsz

        qkv = self.to_qkv(x)
        q, k, v = np.split(qkv, 3, axis = -1)
        q, k, v = map(lambda t: rearrange(t, 'n (h d) -> h n d', h = h), (q, k, v))

        q, k, v = map(lambda t: apply_rotary_pos_emb(t, pos_emb), (q, k, v))
        q, k, v = map(lambda t: rearrange(t, 'h (w n) d -> h w n d', w = window), (q, k, v))

        k, v = map(lambda t: np.pad(t, ((0, 0), (1, 0), (0, 0), (0, 0)), constant_values = 0.), (k ,v))
        k, v = map(lambda t: np.concatenate((t[:, :-1], t[:, 1:]), axis = 2), (k, v))

        sim = np.einsum('h w i d, h w j d -> h w i j', q, k) * self.scale

        mask = np.tril(np.ones((wsz, wsz * 2)), wsz)
        sim = np.where(mask, sim, ATTN_MASK_VALUE)

        sim = sim - np.amax(sim, axis = -1, keepdims = True)
        attn = nn.softmax(sim, axis = -1)

        out = np.einsum('h w i j, h w j d -> h w i d', attn, v)
        out = rearrange(out, 'h w n d -> (w n) (h d)')
        return self.to_out(out)

class FeedForward(hk.Module):
    def __init__(
        self,
        *,
        name,
        dim,
        ff_mult = 4,
        glu = False,
        seq_len = None,
        spatial_gate = False
    ):
        super().__init__(name = name)
        assert not (glu and spatial_gate), 'glu and sgu cannot be turned on at the same time'
        hidden_dim = dim * ff_mult
        hidden_dim *= (1 if not glu else 2)

        self.norm = LayerNorm()
        self.proj_in = hk.Linear(hidden_dim)
        self.proj_out = hk.Linear(dim)

        self.glu = glu
        self.sgu = SGU(dim = hidden_dim, dim_out = hidden_dim // 2, seq_len = seq_len) if spatial_gate else None

    def __call__(self, x):
        x = self.norm(x)
        x = self.proj_in(x)

        if self.glu:
            x, gate = np.split(x, 2, axis = -1)
            x *= nn.gelu(gate)
        else:
            x = nn.gelu(x)

        if exists(self.sgu):
            x = self.sgu(x)

        x = self.proj_out(x)
        return x

class SGU(hk.Module):
    def __init__(
        self,
        *,
        dim,
        dim_out,
        seq_len,
        eps = 1e-3
    ):
        super().__init__()
        self.eps = eps
        self.seq_len = seq_len
        self.norm = LayerNorm()
        self.proj_out = hk.Linear(dim_out)

    def __call__(self, x):
        n = self.seq_len
        x, gate = np.split(x, 2, axis = -1)

        gate = self.norm(gate)

        init_scale = self.eps / n
        init_eps = initializers.RandomUniform(minval = -init_scale, maxval = init_scale)

        weights = hk.get_parameter('spatial_weights', shape = (n, n), init = init_eps)
        biases = hk.get_parameter('spatial_biases', shape = (n, 1), init = np.ones)

        mask = np.tril(np.ones((n, n)))
        weights = weights * mask

        gate = np.einsum('n d, m n -> m d', gate, weights)
        gate += biases

        x = x * gate
        return self.proj_out(x)

class ProGenBase(hk.Module):
    def __init__(
        self,
        *,
        num_tokens,
        dim,
        seq_len,
        depth,
        window_size = 256,
        global_mlp_depth = 2,
        heads = 8,
        dim_head = 64,
        ff_mult = 4,
        ff_glu = True,
        attn_dim = None,
        clamp_gate = True
    ):
        super().__init__()
        self.dim_head = dim_head
        self.embed = hk.Embed(num_tokens, dim)

        self.layers = []
        for i in range(depth):
            use_gmlp = (depth - i) <= global_mlp_depth
            use_ff_glu = not use_gmlp and ff_glu

            self.layers.append([
                LocalAttention(name = f'attn{i}', dim = dim, window_size = window_size, heads = heads, dim_head = dim_head),
                FeedForward(name = f'ff{i}', dim = dim, ff_mult = ff_mult, seq_len = seq_len, spatial_gate = use_gmlp, glu = use_ff_glu)
            ])

        self.to_logits = hk.Sequential([
            LayerNorm(),
            hk.Linear(num_tokens)
        ])

    def __call__(self, x):
        n = x.shape[0]
        x = self.embed(x)
        rotary_emb = fixed_pos_embedding(n, self.dim_head)

        for attn, ff in self.layers:
            x += attn(x, pos_emb = rotary_emb)
            x += ff(x)

        return self.to_logits(x)

def ProGen(mixed_precision = False, mixed_precision_policy = dict(params = 'float32', compute = 'float16', output = 'float32'), **kwargs):
    @hk.transform
    def inner(seq):
        if mixed_precision:
            serialized_policy = ','.join([f'{k}={v}' for k, v in mixed_precision_policy.items()])
            policy = jmp.get_policy(serialized_policy)
            hk.mixed_precision.set_policy(ProGenBase, policy)
        return ProGenBase(**kwargs)(seq)
    return inner



import time
import os, errno
from pathlib import Path
from functools import partial

from google.cloud import storage
from cloudpickle import pickle
from progen_transformer.utils import clear_directory_, silentremove

# filesystem checkpoint fns

def file_reset_checkpoint(path):
    clear_directory_(path)

def file_get_last_checkpoint(path):
    checkpoints = sorted(path.glob('**/ckpt_*'))
    if len(checkpoints) == 0:
        return None

    with open(str(checkpoints[-1]), 'rb') as f:
        package = pickle.load(f)

    return package

def file_save_checkpoint(path, package, keep_last_n = None):
    unix_time = int(time.time())
    checkpoints = sorted(path.glob('**/ckpt_*'))
    num_checkpoints = len(checkpoints)

    with open(str(path / f'ckpt_{unix_time}.pkl'), 'wb') as f:
        pickle.dump(package, f)

    if keep_last_n is None:
        return

    for path_to_rm in checkpoints[:max(0, num_checkpoints - keep_last_n)]:
        silentremove(path_to_rm)

# gcs checkpoint fns

GCS_READ_TIMEOUT = 60 * 30
GCS_WRITE_TIMEOUT = 60 * 30

def gcs_reset_checkpoint(bucket):
    bucket.delete_blobs(list(bucket.list_blobs()))

def gcs_get_last_checkpoint(bucket):
    blobs = sorted(list(bucket.list_blobs()))

    if len(blobs) == 0:
        return None

    last_checkpoint = blobs[-1]

    filename = f'/tmp/{last_checkpoint.name}'
    with open(filename, 'wb') as f:
        last_checkpoint.download_to_file(f, timeout = GCS_READ_TIMEOUT)

    with open(filename, 'rb') as f:
        package = pickle.load(f)

    return package

def gcs_save_checkpoint(bucket, package, keep_last_n = None):
    unix_time = int(time.time())
    blobs = sorted(list(bucket.list_blobs()))
    num_checkpoints = len(blobs)

    filename = f'ckpt_{unix_time}.pkl'
    tmp_path = f'/tmp/{filename}'

    with open(tmp_path, 'wb') as f:
        pickle.dump(package, f)

    blob = bucket.blob(filename)
    blob.upload_from_filename(tmp_path, timeout = GCS_WRITE_TIMEOUT)

    if keep_last_n is None:
        return

    bucket.delete_blobs(blobs[:max(0, num_checkpoints - keep_last_n)])

# factory fn

def get_checkpoint_fns(path):
    use_gcs = path.startswith('gs://')

    if not use_gcs:
        obj = Path(path)
        obj.mkdir(exist_ok = True, parents = True)

        fns = (
            file_reset_checkpoint,
            file_get_last_checkpoint,
            file_save_checkpoint
        )
    else:
        client = storage.Client()
        bucket_name = path[5:]
        obj = client.get_bucket(bucket_name)

        fns = (
            gcs_reset_checkpoint,
            gcs_get_last_checkpoint,
            gcs_save_checkpoint
        )

    fns = tuple(map(lambda fn: partial(fn, obj), fns))
    return fns



from math import ceil
import os, errno
from shutil import rmtree

import jax
from jax import random, nn, value_and_grad, vmap, pmap, jit, lax
from jax.lax import top_k
import jax.numpy as np

from einops import rearrange

# helper functions

def noop(x):
    return x

def exists(val):
    return val is not None

def log(t, eps = 1e-20):
    return np.log(t + eps)

def confirm(question):
    while True:
        resp = input(f'{question} (y/n) ')
        lower_resp = resp.lower()
        if lower_resp in ('y', 'n'):
            return lower_resp == 'y'

def clear_directory_(path):
    rmtree(str(path), ignore_errors = True)
    path.mkdir(exist_ok = True, parents = True)

def silentremove(filename):
    try:
        os.remove(filename)
    except OSError:
        pass

# training functions

def masked_mean(t, mask, axis = None):
    return (t * mask).sum(axis = axis) / mask.sum(axis = axis)

def cross_entropy(logits, targets, axis = -1, ignore_index = 0):
    logprobs = nn.log_softmax(logits, axis = axis)

    nll = np.take_along_axis(logprobs, np.expand_dims(targets, axis = axis), axis = axis)
    nll = nll.squeeze(-1)

    # mask for loss is engineered so that it learns from the first padding token
    # the padding token is reused as end-of-string for simplicity

    mask = (targets != ignore_index)
    eos_mask = (~mask).cumsum(axis = -1) == 1
    mask = mask | eos_mask

    ce = -masked_mean(nll, mask, axis = -1)
    return ce

def get_loss_fn(model, data_parallel = False):
    def loss_fn(params, key, data):
        ids, labels = data[:-1], data[1:]
        logits = model.apply(params, key, ids)
        return cross_entropy(logits, labels, axis = -1)

    loss_fn = jit(vmap(loss_fn, in_axes = (None, None, 0), out_axes = 0))

    if data_parallel:
        loss_fn = pmap(loss_fn, in_axes = (None, None, 0), out_axes = 0)

    @value_and_grad
    def batched_loss_fn(params, key, data):
        if not data_parallel:
            values = loss_fn(params, key, data)
            return np.mean(values)

        mask = np.ones((data.shape[0],))

        device_count = jax.local_device_count()
        batch_size = data.shape[0]

        remainder = (batch_size % device_count)
        if remainder != 0:
            padding = device_count - remainder
            data = np.pad(data, ((0, padding), (0, 0)))
            mask = np.pad(mask, ((0, padding)))

        data, mask = map(lambda t: rearrange(t, '(p b) ... -> p b ...', p = device_count), (data, mask))
        values = loss_fn(params, key, data)
        return masked_mean(values, mask)

    return batched_loss_fn

# sampling functions

def select_top_k(tensor, k):
    values, _ = top_k(tensor, k)
    mask = tensor > values.min()
    return mask, np.where(mask, tensor, 0.)

def gumbel_noise(rng, shape):
    noise = random.uniform(rng, shape = shape, minval = 0., maxval = 1.)
    return -log(-log(noise))

def sample(rng, fn, params, prime, length, top_k = None, add_bos = False):
    start_pos = prime.shape[-1]
    pad_right = length - prime.shape[-1]

    padding = (0, pad_right) if not add_bos else (1, pad_right - 1)
    seq = np.pad(prime, padding)

    one_hots = np.eye(length, dtype = int)

    for curr_pos in range(start_pos, length):
        logits = fn(params, next(rng), seq)
        logits = logits[curr_pos - 1]

        noise = gumbel_noise(next(rng), logits.shape)

        if exists(top_k):
            mask, logits = select_top_k(logits, top_k)
            noise *= mask

        logits += noise
        sampled_ind = np.argmax(logits, axis = -1)

        one_hot = one_hots[curr_pos]
        seq += one_hot * sampled_ind

    # for now, just set everything after second padding token (eos) to padding
    remove_after_eos_mask = (seq == 0).cumsum(axis = -1) > 1
    seq *= ~remove_after_eos_mask

    return seq

# rng hacks

def hardware_uniform(
    rng_key,
    shape,
    dtype = np.float32,
    minval = np.float32(0),
    maxval = np.float32(1)
):
    del rng_key
    minval = lax.convert_element_type(minval, dtype)
    maxval = lax.convert_element_type(maxval, dtype)
    return lax.rng_uniform(minval, maxval, shape)

def hardware_bernoulli(rng_key, p = np.float32(0.5), shape = None):
    del rng_key
    return lax.rng_uniform(0.0, 1.0, shape) < p

def set_hardware_rng_(jax):
    jax.random.bernoulli = hardware_bernoulli
    jax.random.uniform = hardware_uniform
    jax._src.random.uniform = hardware_uniform



from setuptools import setup, find_packages

setup(
    name="progen-transformer",
    packages=find_packages(),
    version="0.0.35",
    license="MIT",
    description="Protein Generation (ProGen)",
    author="Phil Wang",
    author_email="",
    url="https://github.com/lucidrains/progen",
    keywords=[
        "artificial intelligence",
        "deep learning",
        "protein language model"
    ],
    install_requires=[
        "click",
        "click-option-group",
        "cloudpickle",
        "einops>=0.3",
        "dagster",
        "dagit",
        "dm-haiku",
        "google-cloud-storage",
        "humanize",
        "jax",
        "jaxlib",
        "Jinja2",
        "jmp",
        "omegaconf",
        "optax>=0.0.9",
        "python-dotenv",
        "pyfaidx",
        "tensorflow",
        "tqdm",
        "wandb"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
    ],
)



import tensorflow as tf
import numpy as np
from functools import partial
from pathlib import Path
from contextlib import contextmanager

# writing tfrecords

def write(writer, values):
    record_bytes = tf.train.Example(features = tf.train.Features(feature={
        'seq': tf.train.Feature(bytes_list = tf.train.BytesList(value=[values]))
    })).SerializeToString()

    writer.write(record_bytes)

@contextmanager
def with_tfrecord_writer(path):
    options = tf.io.TFRecordOptions(compression_type = 'GZIP')

    with tf.io.TFRecordWriter(path, options = options) as writer:
        yield partial(write, writer)

# reading tfrecords

def parse_fn(sample):
    return tf.io.parse_single_example(sample, {
        'seq': tf.io.FixedLenFeature([], tf.string)
    })

def collate_fn(batch, pad_length, offset = 0):
    tensors = [np.frombuffer(el, dtype = np.uint8).astype(np.uint16) for el in batch.numpy()]
    tensors = map(lambda t: t[..., :pad_length], tensors)
    tensors = map(lambda t: t + offset, tensors)
    padded_tensors = map(lambda t: np.pad(t, (0, pad_length - t.shape[-1])), tensors)
    return np.stack(list(padded_tensors))

def iterator_from_tfrecords_folder(folder, data_type = 'train'):
    is_gcs_path = folder.startswith('gs://')

    if is_gcs_path:
        filenames = tf.io.gfile.glob(f'{folder}/*.{data_type}.tfrecord.gz')
    else:
        folder = Path(folder)
        filenames = [str(p) for p in folder.glob(f'**/*.{data_type}.tfrecord.gz')]

    num_seqs = sum(map(lambda t: int(t.split('.')[-4]), filenames))

    def iter_fn(
        seq_len,
        batch_size,
        skip = 0,
        loop = False
    ):
        dataset = tf.data.TFRecordDataset(filenames, compression_type = 'GZIP')

        dataset = dataset.skip(skip)
        dataset = dataset.map(parse_fn)
        dataset = dataset.batch(batch_size)
        dataset = dataset.prefetch(tf.data.AUTOTUNE)

        if loop:
            dataset = dataset.repeat()

        for batch in dataset:
            seq = batch['seq']
            batch_size = seq.shape[0]
            seq = collate_fn(seq, pad_length = seq_len, offset = 1)
            bos = np.zeros((batch_size, 1), dtype = np.uint16)
            seq = np.concatenate((bos, seq), axis = 1)
            yield seq

    return num_seqs, iter_fn

# tokenization

def encode_token(token):
    return ord(token) + 1

def decode_token(token):
    if token < 0:
        return ''
    return str(chr(token))

def encode_tokens(tokens):
    return list(map(encode_token, tokens))

def decode_tokens(tokens, offset = 1):
    return ''.join(list(map(decode_token, tokens.astype(np.int16) - offset)))
